from ..base import AbstractValidator


class TitleValidator(AbstractValidator):
    name = 'Проверка полей заголовка'
    code = '2'

    def __init__(self, schema):
        self._schema = schema
        self.errors = []

        self._report_fields = []
        self._schema_fields = self._get_schema_fields()

    def __repr__(self):
        return ('<TitleValidator schema_fields={_schema_fields} '
                'report_fields={_report_fields} errors={errors}>').format(
                    **self.__dict__)

    def _get_schema_fields(self):
        '''Создание множества с именами полей заголовка'''
        return set(f for f in self._schema.title.xpath('./item/@field'))

    def validate(self, report):
        self._check_common(report)
        self._check_required_fields()
        self._check_missing_fields()

        return not bool(self.errors)

    def _check_common(self, report):
        '''Выполенние цикла опервичных проверок'''
        for field, value in report.title:
            self.__check_extra(field)
            self.__check_dup(field)
            self.__check_value(field, value)
            self.__check_okpo(field, value)

            self._report_fields.append(field)

    def __check_extra(self, field):
        '''Проверка, является ли поле лишним'''
        if field not in self._schema_fields:
            self.error(f'Лишнее поле - {field}', '1')

    def __check_dup(self, field):
        '''Проверка, является ли поле дубликатом'''
        if self._report_fields.count(field) != 0:
            self.error(f'Повтор поля - {field}', '2')

    def __check_value(self, field, value):
        '''Проверка значения в поле'''
        if not value:
            self.error(f'Отсутствует значение в поле - {field}', '3')

    def __check_okpo(self, field, value):
        '''Проверка формата ОКПО'''
        def __is_valid_okpo():
            if len(value) in (8, 10, 14) and value.isdigit():
                return True
            return False

        if field == self._schema.obj and not __is_valid_okpo():
            self.error('Код ОКПО должен быть длиной 8, 10 или 14 цифр', '4')

    def _check_required_fields(self):
        '''Проверка наличия ключевого поля в заголовке'''
        if self._schema.obj not in self._report_fields:
            self._schema_fields.discard(self._schema.obj)
            self.error(f'Отсутствует ключевое поле - {self._schema.obj}', '5')

    def _check_missing_fields(self):
        '''Проверка на отсутствие в отчёте полей, описанных в схеме'''
        for filed in self._schema_fields - set(self._report_fields):
            self.error(f'Отсутствует поле - {filed}', '6')



class ControlError(Exception):
    pass


class PeriodExprError(ControlError):
    '''Ошибка разбора формулы проверки периодичности'''


class ConditionExprError(ControlError):
    '''Ошибка разбора условия контроля'''


class RuleExprError(ControlError):
    '''Ошибка разбора правила контроля'''


class PrevPeriodNotImpl(ControlError):
    def __init__(self, id):
        self.id = id
        self.msg = 'Проверка со значениями из прошлого периода не реализована'



class FormatError(Exception):
    pass


# ---


class NoSectionReportError(FormatError):
    def __init__(self, sec_code):
        self.code = '1'
        self.msg = 'Раздел {} отсутствует в отчёте'.format(sec_code)


class DuplicateError(FormatError):
    def __init__(self, row_code, counter):
        self.code = '2'
        self.msg = 'Строка {} повторяется {} раз(а)'.format(row_code, counter)


class EmptyRowError(FormatError):
    def __init__(self, sec_code, row_code):
        self.code = '3'
        self.msg = 'Раздел {}, строка {} не может быть пустой'.format(sec_code,
                                                                      row_code)


class EmptyColumnError(FormatError):
    def __init__(self, sec_code, row_code, col_code):
        self.code = '4'
        self.msg = ('Раздел {}, строка {}, графа {} не может быть пустой'
                    .format(sec_code, row_code, col_code))


class NoSectionTemplateError(FormatError):
    def __init__(self, sec_code):
        self.code = '5'
        self.msg = 'Раздел {} не описан в шаблоне'.format(sec_code)


class NoRuleError(FormatError):
    def __init__(self, sec_code, row_code, col_code):
        self.code = '6'
        self.msg = ('Раздел {}, строка {}, графа {}. '
                    'В шаблоне отсутствует правило для проверки этого поля'
                    .format(sec_code, row_code, col_code))


# ---


class SpecBaseError(FormatError):
    def update(self, coords, spec):
        self.msg = 'Раздел {}, строка {}, специфика {}. {}'.format(coords[0],
                                                                   coords[1],
                                                                   spec,
                                                                   self.msg)


class SpecNotInDictError(SpecBaseError):
    msg = 'Специфика отсутствует в справочнике'
    code = '7'


class SpecValueError(SpecBaseError):
    msg = 'Недопустмое значение'
    code = '8'


# ---


class ValueBaseError(FormatError):
    def update(self, coords):
        self.msg = 'Раздел {}, строка {}, графа {}. {}'.format(*coords,
                                                               self.msg)


class ValueNotNumberError(ValueBaseError):
    msg = 'Значение не является числом'
    code = '9'


class ValueBadFormat(ValueBaseError):
    msg = 'Число не соответствует формату'
    code = '10'


class ValueLengthError(ValueBaseError):
    msg = 'Длина строки больше допустимого'
    code = '11'


class ValueNotInDictError(ValueBaseError):
    msg = 'Значение отсутствует в справочнике'
    code = '12'


class ValueNotInRangeError(ValueBaseError):
    msg = 'Значение не входит в диапазон допустимых'
    code = '13'


class ValueNotInListError(ValueBaseError):
    msg = 'Значение не входит в список допустимых'
    code = '14'



import operator
from math import floor
from copy import deepcopy
from itertools import chain
from functools import reduce

operator_map = {
    '<': operator.lt,
    '<=': operator.le,
    '=': operator.eq,
    '>': operator.gt,
    '>=': operator.ge,
    '<>': operator.ne,
    'and': operator.and_,
    'or': operator.or_,
    '+': operator.add,
    '-': operator.sub,
    '*': operator.mul,
    '/': operator.truediv,
}


class Elem:
    def __init__(self, val, section=[], rows=[], columns=[]):
        self.section = set(section)
        self.rows = set(rows)
        self.columns = set(columns)

        self._controls = []
        self._func = None

        self.bool = True
        self.val = None if val is None else float(val)

    def __add__(self, elem):
        return self.__modify(elem, operator.add)

    def __sub__(self, elem):
        return self.__modify(elem, operator.sub)

    def __mul__(self, elem):
        return self.__modify(elem, operator.mul)

    def __truediv__(self, elem):
        return self.__modify(elem, operator.truediv)

    def __neg__(self):
        self.val = -self.val
        return self

    def __repr__(self):
        return ('<Elem {}{}{} value={} bool={}>').format(
            sorted(int(i) for i in self.section),
            sorted(int(i) for i in self.rows if i.isdigit()),
            sorted(int(i) for i in self.columns if i.isdigit()),
            self.val, self.bool)

    def __modify(self, elem, op_func):
        self.rows |= elem.rows
        self.columns |= elem.columns
        try:
            self.val = op_func(self.val, elem.val)
        except ZeroDivisionError:
            pass
        return self

    @property
    def controls(self):
        return self._controls

    def controls_clear(self):
        '''Очищение списка контролей'''
        self._controls.clear()

    def controls_extend(self, r_elem):
        '''Расширение списка контролей, контролями из правого элемента'''
        self._controls.extend(r_elem._controls)

    def controls_append(self, r_elem, op_name):
        '''Форматирование и добавление непройденного контроля'''
        self._controls.append({
            'left': self.val,
            'operator': op_name,
            'right': r_elem.val,
            'delta': round(self.val - r_elem.val, 2)
        })

    def check(self, report, params, ctx_elem):
        if self._func:
            return self._apply_func(report, params, *self._func)
        return [self]

    def _apply_func(self, report, params, func, right_elem):
        '''Выполнение функции на элементах массива'''
        elems = []
        for r_elem in right_elem.check(report, params, self):
            elems.append(getattr(operator, func)(self, r_elem))
        return elems

    def isnull(self, replace):
        '''Замена None на replace. Так же снимает признак "заглушки"'''
        self.val = float(replace) if self.val is None else self.val

    def round(self, ndig, trunc=0):
        '''Округление/отсечение до ndig знаков'''
        if trunc > 0:
            self.val = float(f'{self.val:.{abs(ndig)}f}')
        else:
            self.val = round(self.val, ndig)

    def abs(self):
        '''Выполнение функции abs над значением'''
        self.val = abs(self.val)

    def floor(self):
        '''Выполнение функции floor над значением'''
        self.val = floor(self.val)

    def add_func(self, func, arg):
        '''Добавляем функцию элементу при парсинге'''
        self._func = (func, arg)


class ElemList:
    def __init__(self, section, rows, columns,
                 s1=[None], s2=[None], s3=[None]):
        self.section = section[0]
        self.rows = set(rows)
        self.columns = set(columns)

        self.specs = {1: s1, 2: s2, 3: s3}

        self.funcs = []
        self.elems = []

    def __repr__(self):
        return '<ElemList [{}]{}{} funcs={} elems={}>'.format(
            self.section,
            sorted(int(i) for i in self.rows if i.isdigit()),
            sorted(int(i) for i in self.columns if i.isdigit()),
            self.funcs, self.elems)

    def __neg__(self):
        self._apply_unary('neg')
        return self

    def check(self, report, params, ctx_elem):
        self._prepare_specs(params.formats, params.catalogs)
        self._read_data(report, params.dimension)
        self._apply_funcs(report, params, ctx_elem)
        return self._flatten_elems()

    def _prepare_specs(self, formats, catalogs):
        '''Подготовка специфик. Если список специфик не "пуст" и не содержит
           один только ключ "*", пытаемся "развернуть" его к простому списку.
           Другими словами, все диапазоны специфик привести к явному набору в
           соответствии со справочником. Затем конвертируем всё в множество
        '''
        for spec_idx in range(1, 4):
            if self.specs[spec_idx] in ([None], ['*']):
                self.specs[spec_idx] = set(self.specs[spec_idx])
            else:
                spec_list = self.__get_spec_list(formats, catalogs, spec_idx)
                self.specs[spec_idx] = set(self.__expand_specs(spec_list,
                                                               spec_idx))

    def __get_spec_list(self, formats, catalogs, spec_idx):
        '''Определяем набор параметров для специфики указанной строки, раздела.
           Выбираем список специфик по имени справочника из наборапараметров
        '''
        row_code = next(iter(self.rows))
        params = formats.get_spec_params(self.section, row_code, spec_idx)
        return catalogs.get(params.get('dic'), {}).get('ids', [])

    def __expand_specs(self, spec_list, spec_idx):
        '''Перебираем специфики. Простые специфики сразу возвращаем. Если
           имеем диапазон, определяем индекс начальной и конечной специфик
           из списка-справочника, итерируемся по определенному диапазону,
           возвращая соответствующие специфики из списка-справочника
        '''
        for spec in self.specs[spec_idx]:
            if '-' in spec:
                start, end = spec.split('-')
                for i in range(spec_list.index(start.strip()),
                               spec_list.index(end.strip()) + 1):
                    yield spec_list[i]
            else:
                yield spec

    def _read_data(self, report, dimension):
        '''Чтение отчёта и конвертация его в массивы элементов'''
        raw_sec = report.get_section(self.section)
        for row_code, raw_rows in self._read_rows(raw_sec):
            if not raw_rows:
                self.elems.append(self._proc_row_empty(row_code, dimension))
                continue
            for raw_row in raw_rows:
                self.elems.append(self._proc_row(raw_row, row_code, dimension))

    def _read_rows(self, raw_sec):
        '''Читаем строки'''
        if self.rows == {'*'}:
            return raw_sec.items(specs=self.specs)
        return raw_sec.items(codes=self.rows, specs=self.specs)

    def _read_columns(self, raw_row, dimension):
        '''Читаем графы если строка не пустая'''
        if self.columns == {'*'}:
            return raw_row.items(codes=dimension[self.section])
        return raw_row.items(codes=self.columns)

    def _read_columns_empty(self, dimension):
        '''Читаем графы если строка пустая'''
        if self.columns == {'*'}:
            return iter(dimension[self.section])
        return iter(self.columns)

    def _proc_row(self, raw_row, row_code, dimension):
        '''Заполняем строку элементами со значениями из отчета, отсутствующие
           значения замещаем заглушкой
        '''
        row = []
        for col_code, value in self._read_columns(raw_row, dimension):
            row.append(Elem(value or 0, self.section, [row_code], [col_code]))
        return row

    def _proc_row_empty(self, row_code, dimension):
        '''Заполняем строку элементами заглушками'''
        row = []
        for col_code in self._read_columns_empty(dimension):
            row.append(Elem(0, self.section, [row_code], [col_code]))
        return row

    def _apply_funcs(self, report, params, ctx_elem):
        '''Выполнение функций на эелементах массива'''
        for func, args in self.funcs:
            if func == 'sum':
                self._apply_sum(ctx_elem)
            elif func in ('abs', 'floor'):
                self._apply_unary(func)
            elif func in ('round', 'isnull'):
                self._apply_binary(report, params, func, args)
            else:
                self._apply_math(report, params, func, *args)

    def _apply_sum(self, ctx_elem):
        '''Суммирование строк и/или графов'''
        if self.columns == ctx_elem.columns:  # строк в каждой графе
            self.elems = [[reduce(operator.add, l)] for l in zip(*self.elems)]
        elif self.rows == ctx_elem.rows:      # граф в каждой строке
            self.elems = [[reduce(operator.add, l)] for l in self.elems]
        elif not self.elems:                  # всех ячеек (секция пустая)
            self.elems = [[Elem(0, self.section, '*', '*')]]
        else:                                 # всех ячеек (секция не пустая)
            self.elems = [[reduce(operator.add, chain(*self.elems))]]

    def _apply_unary(self, func):
        '''Выполнение унарных операций (abs, floor, neg)'''
        for row in self.elems:
            for elem in row:
                getattr(elem, func)()

    def _apply_binary(self, report, params, func, elems):
        '''Выполнение бинарных операций (round, isnull)'''
        args = [int(elem.check(report, params, self)[0].val) for elem in elems]
        for row in self.elems:
            for elem in row:
                getattr(elem, func)(*args)

    def _apply_math(self, report, params, func, elem):
        '''Выполнение математических операций (add, sub, mul, truediv)'''
        left_operand = self._flatten_elems()
        right_operand = elem.check(report, params, self)

        operand_pairs = self._zip(left_operand, right_operand)
        self.elems.clear()
        for l_elem, r_elem in operand_pairs:
            self.elems.append([getattr(operator, func)(l_elem, r_elem)])

    def _zip(self, *lists):
        '''Сбираем массивы в список кортежей. Если длина массивов различается
           и самый короткий длиной в 1 элемент, заменяем его на массив равной
           длины с полными копиями этого элемента
           (Предполагается, что только 1 массив может быть короче других и
           его длина всегда равна 1, но проверку на всякий случай оставил)
           [1, 2], [3], [4, 5] > [1, 2], [3, 3], [4, 5] > [1, 3, 4], [2, 3, 5]
        '''
        smallest, *_, biggest = sorted(lists, key=len)
        if len(smallest) != len(biggest) and len(smallest) == 1:
            lists = list(lists)
            s_idx, s_elem = lists.index(smallest), smallest[0]
            lists[s_idx] = [deepcopy(s_elem) for _ in range(len(biggest))]
        return zip(*lists)

    def _flatten_elems(self):
        '''Возвращаем плоский массив элементов'''
        return list(chain(*self.elems))

    def add_func(self, func, *args):
        '''Добавляем функцию в "очередь" при парсинге'''
        self.funcs.append((func, args))


class ElemLogic(ElemList):
    def __init__(self, l_elem, operator, r_elem):
        self.l_elem = l_elem
        self.r_elem = r_elem
        self.op_name = operator.lower()
        self.op_func = operator_map[self.op_name]

        self.elems = []

        self.params = None

    def __repr__(self):
        return '<ElemLogic left={} operator="{}" right={}>'.format(
            self.l_elem, self.op_name, self.r_elem)

    def check(self, report, params, ctx_elem=None):
        '''Основной метод вызова проверки'''
        self.params = params
        self._control(report)
        return self.elems

    def _control(self, report):
        '''Подготовка элементов, слияние, передача в метод контроля'''
        l_elems = self.l_elem.check(report, self.params, self.r_elem)
        r_elems = self.r_elem.check(report, self.params, self.l_elem)
        elems_pairs = self._zip(l_elems, r_elems)

        self.__control(elems_pairs)

    def __control(self, elems_pairs):
        '''Определение аттрибута контроля. Итерация по парам элементов,
           выполнение проверок, обработка результата
        '''
        attrib = 'bool' if self.op_name in ('and', 'or') else 'val'

        for l_elem, r_elem in elems_pairs:
            if not self.__logic_control(l_elem, r_elem, attrib):
                self.__get_result(l_elem, r_elem, success=False)
            else:
                self.__get_result(l_elem, r_elem, success=True)

            self.elems.append(l_elem)

    def __get_result(self, l_elem, r_elem, *, success):
        '''Обработка результата. При проверке логического "or", если левый
           элемент уже содержит ошибки, затираем их, инчае затираем список
           контролей правого. При неуспешной проверке, формируется ошибка
           которая прибавляется к списку контролей левого элемента.
           Затем в левый элемент сливаются все ошибки из списка правого.
        '''
        if self.op_name == 'or':
            if l_elem.controls:
                l_elem.controls_clear()
            else:
                r_elem.controls_clear()

        if not success:
            l_elem.controls_append(r_elem, self.op_name)
        l_elem.val = r_elem.val
        l_elem.controls_extend(r_elem)

    def __logic_control(self, l_elem, r_elem, attrib):
        '''Получение значений и проведение проверки'''
        l_elem_v, r_elem_v = self.__get_elem_values(l_elem, r_elem, attrib)
        if not self.op_func(l_elem_v, r_elem_v):
            return self.__check_fault(l_elem.val, r_elem.val)
        return True

    def __get_elem_values(self, l_elem, r_elem, attrib):
        '''Округление и возвращение значений которые будут сравниваться'''
        l_elem.round(self.params.precision)
        r_elem.round(self.params.precision)
        return getattr(l_elem, attrib), getattr(r_elem, attrib)

    def __check_fault(self, l_elem_v, r_elem_v):
        '''Проверка погрешности'''
        return abs(l_elem_v - r_elem_v) <= self.params.fault


class ElemSelector(ElemList):
    def __init__(self, action, elems):
        self.action = action.lower()
        self.funcs = []
        self.elems = elems

    def __repr__(self):
        return '<ElemSelector action={} funcs={} elems={}>'.format(
            self.action, self.funcs, self.elems)

    def check(self, *args):
        self._select(args)
        self._apply_funcs(*args)
        return self._flatten_elems()

    def _select(self, args):
        '''Подготовка элементов, слияние. Очистка списка элементов.
           Вызов метода селектора по полю action
        '''
        elems_results = self._zip(*(elem.check(*args) for elem in self.elems))
        self.elems.clear()
        getattr(self, self.action)(elems_results)

    def nullif(self, elems_results):
        '''Сравнивает результаты левого и правого элементов. Добавляем к
           результату элемент со значением None если значения равны,
           иначе добавляем левый элемент
        '''
        for l_elem, r_elem in elems_results:
            if l_elem.val == r_elem.val:
                self.elems.append([Elem(0)])
            else:
                self.elems.append([l_elem])

    def coalesce(self, elems_results):
        '''Сравнивает результаты элементов каждой "линии" (строки/графа).
           Добавляем к результату первый элемент значение которого не None
        '''
        for line_elems in elems_results:
            first_elem = next([e] for e in line_elems if e.val is not None)
            self.elems.append(first_elem)



from re import IGNORECASE, DOTALL
import ply.lex as lex
from ply.lex import TOKEN

tokens = ('COMMA', 'PLUS', 'MINUS', 'MULTIPLY', 'DIVIDE', 'LPAREN', 'RPAREN',
          'ELEM_START', 'ELEM_END', 'CODE', 'SUM', 'ABS', 'FLOOR', 'ROUND',
          'ISNULL', 'NULLIF', 'COALESCE', 'LOGIC', 'NUM', 'COMP')

t_ignore = ' \r\t\f'

t_COMMA = r','
t_PLUS = r'\+'
t_MINUS = r'-'
t_DIVIDE = r'/'
t_MULTIPLY = r'\*'
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_ELEM_START = r'{?{'
t_ELEM_END = r'}?}'

t_LOGIC = r'and|or'
t_COMP = r'\| *(<|<=|=|>=|>|<>) *\|'

t_SUM = r'SUM'
t_ABS = r'abs'
t_FLOOR = r'floor'
t_ROUND = r'round'
t_ISNULL = r'isnull'
t_NULLIF = r'nullif'
t_COALESCE = r'coalesce'


def _range(rng):
    start, end = rng.split('-')
    return (str(i) for i in range(int(start), int(end) + 1))


@TOKEN(r'\[.+?\]')
def t_CODE(t):
    code = []
    for i in map(lambda i: i.strip(), t.value[1:-1].split(',')):
        if ('-' in i) and ('.' not in i):
            code.extend(_range(i))
        else:
            code.append(i)
    t.value = code
    return t


@TOKEN(r'\d+(\.\d+)?')
def t_NUM(t):
    t.value = float(t.value)
    return t


@TOKEN(r'\n+')
def t_newline(t):
    t.lexer.lineno += len(t.value)


def t_error(t):
    print("Illegal character '%s'" % t.value[0])
    t.lexer.skip(1)


lexer = lex.lex(reflags=IGNORECASE | DOTALL)



from ..exceptions import SpecBaseError, SpecNotInDictError, SpecValueError


class SpecInspector:
    def __init__(self, params, catalogs):
        self._catalogs = catalogs

        self.catalog = params.get('dic')
        self.vld_type = params.get('vldType')
        self.vld_param = params.get('vld')

    def __repr__(self):
        return ('<SpecInspector vld_type={vld_type} '
                'vld_param={vld_param}>').format(**self.__dict__)

    def check(self, coords, row, spec_idx, specs_map):
        try:
            self._check(row, spec_idx, specs_map)
        except SpecBaseError as ex:
            ex.update(coords, spec_idx)
            raise

    def _check(self, row, spec_idx, specs_map):
        if self.vld_type == '4':
            self.__check_value_catalog_add(row, spec_idx)
        elif self.vld_type == '5':
            self.__check_value_catalog_coord(row, spec_idx, specs_map)

    def __check_value_catalog_add(self, row, spec_idx):
        '''Проверка на вхождение в справочник'''
        if row.get_spec(spec_idx) not in self._catalogs[self.vld_param]['ids']:
            raise SpecNotInDictError()

    def __check_value_catalog_coord(self, row, spec_idx, specs_map):
        '''Проверка на вхождение в справочник и связь с главной спецификой'''
        catalog, coords = self.vld_param.split('=#')
        *_, col_code = coords.split(',')

        spec = row.get_spec(spec_idx)
        ctx_spec = row.get_spec(specs_map[col_code])

        try:
            ctx_catalog = self._catalogs[self.catalog]['full'][spec][catalog]
            if ctx_spec not in ctx_catalog:
                raise SpecValueError()
        except KeyError:
            raise SpecValueError()



from itertools import chain
from collections import namedtuple
from ..parser import parser
from ..exceptions import ConditionExprError, RuleExprError, PrevPeriodNotImpl


ControlParams = namedtuple('ControlParams', ('is_rule',
                                             'formats',
                                             'catalogs',
                                             'dimension',
                                             'precision',
                                             'fault'))


class FormulaInspector:
    def __init__(self, control, *, formats, catalogs, dimension, skip_warns):
        self._skip_warns = skip_warns

        self.formats = formats
        self.catalogs = catalogs
        self.dimension = dimension
        self.id = control.attrib['id']
        self.name = control.attrib['name']
        self.rule = control.attrib['rule'].strip()
        self.condition = control.attrib['condition'].strip()

        self.tip = int(control.attrib.get('tip', '1'))
        self.fault = float(control.attrib.get('fault', '-1'))
        self.precision = int(control.attrib.get('precision', '2'))

    def __repr__(self):
        return ('<FormulaInspector id={id} name={name} rule={rule} '
                'condition={condition} fault={fault} '
                'precision={precision}>').format(**self.__dict__)

    def check(self, report):
        if self._check_condition(report):
            return self._check_rule(report)
        return []

    def _check_condition(self, report):
        '''Проверка условия для выполнения контроля'''
        if self.condition and not self._is_previous_period(self.condition):
            evaluator = self.__parse(self.condition, ConditionExprError)
            return not list(self.__check(report, evaluator, self.__params()))
        return True

    def _check_rule(self, report):
        '''Проверка правила контроля'''
        if self.rule and not self._is_previous_period(self.rule):
            evaluator = self.__parse(self.rule, RuleExprError)
            return self.__check(report, evaluator, self.__params(is_rule=True))
        return []

    def __params(self, is_rule=False):
        '''Упаковка параметров для проверки в именованный кортеж'''
        return ControlParams(is_rule,
                             self.formats,
                             self.catalogs,
                             self.dimension,
                             self.precision,
                             self.fault if is_rule else float(-1))

    def __parse(self, formula, exc):
        '''Парсинг формулы контроля'''
        evaluator = parser.parse(formula)
        if evaluator is None:
            raise exc(self.id)
        return evaluator

    def __check(self, report, evaluator, params):
        '''Выполнение проверки. Возвращает список проваленых проверок'''
        results = evaluator.check(report, params)
        return chain.from_iterable(result.controls for result in results)

    def _is_previous_period(self, formula):
        '''Проверка наличия в формуле элемента в двух фигурных скобках,
           что говорит о том, что значение берётся за прошлый период.
           Такой функционал пока неизвестно когда получится реализовать
        '''
        if '{{' in formula:
            if self._skip_warns:
                return True
            else:
                raise PrevPeriodNotImpl(self.id)



from os.path import isfile
from io import BytesIO, BufferedIOBase
from lxml import etree
from .report import Report
from .schema import Schema


def _get_xml_etree(source):
    if isinstance(source, (etree._ElementTree, etree._Element)):
        return source
    elif isinstance(source, str) and isfile(source):
        return etree.parse(source)
    elif isinstance(source, bytes):
        return etree.parse(BytesIO(source))
    elif isinstance(source, BufferedIOBase):
        return etree.parse(source)

    raise TypeError(f'Expected ElementTree, Element, bytes, file name/path, '
                    f'or file-like object, got {source!r}')


def parse_report(source):
    xml_etree = _get_xml_etree(source)
    return Report(xml_etree)


def parse_schema(source, skip_warns=False):
    xml_etree = _get_xml_etree(source)
    return Schema(xml_etree, skip_warns=skip_warns)



from ..base import AbstractValidator
from .exceptions import PrevPeriodNotImpl
from .inspectors import PeriodInspector, FormulaInspector


class ControlValidator(AbstractValidator):
    name = 'Проверка контролей'
    code = '4'

    def __init__(self, schema):
        self._schema = schema
        self.errors = []

        self._template = ('{control_name}; слева {left} {operator} '
                          'справа {right} разница {delta}')

    def __repr__(self):
        return '<ControlValidator errors={errors}>'.format(**self.__dict__)

    def __fmt_control(self, ctrl, name):
        '''Форматирование сообщения о непройденном контроле'''
        return self._template.format(control_name=name, **ctrl)

    def validate(self, report):
        self._check_controls(report)

        return not bool(self.errors)

    def _check_controls(self, report):
        '''Проверка отчёта по контролям'''
        if report.blank:
            return

        for control in self._schema.controls:
            self._check_control(report, control)

    def _check_control(self, report, control):
        '''Обёртка для обработки исключения'''
        try:
            if self.__check_period(report, control):
                self.__check_control(report, control)
        except PrevPeriodNotImpl as ex:
            self.error(ex.msg, ex.id, level=0)

    def __check_period(self, report, control):
        '''Проверка соответствия периода контроля периоду в отчёте'''
        inspector = PeriodInspector(control)
        return inspector.check(report)

    def __check_control(self, report, control):
        '''Проверка контрольных значений отчёта'''
        inspector = FormulaInspector(control,
                                     formats=self._schema.formats,
                                     catalogs=self._schema.catalogs,
                                     dimension=self._schema.dimension,
                                     skip_warns=self._schema.skip_warns)
        for ctrl in inspector.check(report):
            message = self.__fmt_control(ctrl, inspector.name)
            self.error(message, inspector.id, level=inspector.tip)



from collections import defaultdict


class SchemaFormats(dict):
    def _get_spec_code(self, sec_code, spec_idx):
        '''Возвращает из указаной секции ключ специфики по её индексу'''
        for code, idx in self[sec_code]['specs'].items():
            if idx == spec_idx:
                return code

    def get_spec_params(self, sec_code, row_code, spec_idx):
        '''Возвращает для указанной секции и строки словарь параметров,
           определяющий формат проверок для специфики с указанным кодом
        '''
        spec_code = self._get_spec_code(sec_code, str(spec_idx))
        return self[sec_code][row_code][spec_code]


class NestedDefaultdict(dict):
    def __init__(self, default_factory):
        self.default_factory = default_factory

    def __repr__(self):
        return '<nesteddefaultdict {}>'.format(super().__repr__())

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            self[key] = defaultdict(self.default_factory)
            return self[key]



from ..base import AbstractValidator
from .inspectors import ValueInspector, SpecInspector
from .exceptions import (FormatError, DuplicateError, EmptyRowError,
                         EmptyColumnError, NoSectionReportError,
                         NoSectionTemplateError, NoRuleError)


class FormatValidator(AbstractValidator):
    name = 'Проверка формата'
    code = '3'

    def __init__(self, schema):
        self._schema = schema
        self.errors = []

    def __repr__(self):
        return '<FormatValidator errors={errors}>'.format(**self.__dict__)

    def validate(self, report):
        try:
            self._check_sections(report)
            self._check_duplicates(report)
            self._check_required(report)
            self._check_format(report)
        except FormatError as ex:
            self.error(ex.msg, ex.code)

        return not bool(self.errors)

    def _check_sections(self, report):
        '''Проверка целостности отчёта'''
        report_sections = set(code for code, _ in report.items())
        schema_sections = set(self._schema.dimension.keys())

        for section in schema_sections - report_sections:
            raise NoSectionReportError(section)

    def _check_duplicates(self, report):
        '''Проверка дубликатов строк'''
        def __fmt_specs(specs):
            return ' '.join(f's{i}={s}' for i, s in enumerate(specs, 1) if s)

        for row, counter in report.row_counters.items():
            if counter > 1:
                row_code, *specs = row
                if any(specs):
                    row = f'{row_code} {__fmt_specs(specs)}'
                else:
                    row = row_code
                raise DuplicateError(row, counter)

    def _check_required(self, report):
        '''Проверка наличия обязательных к заполнению строк и значений'''
        for sec_code, row_code, col_code in self._schema.required:
            rows = list(report.get_section(sec_code).get_rows(row_code))
            if not rows:
                raise EmptyRowError(sec_code, row_code)

            for row in rows:
                if not row.get_col(col_code):
                    raise EmptyColumnError(sec_code, row_code, col_code)

    def _check_format(self, report):
        '''Проверка формата строк и значений в них'''
        for sec_code, section in report.items():
            for row_code, rows in section.items():
                for row in rows:
                    self.__check_row(sec_code, row_code, row)
                    self.__check_cells(sec_code, row_code, row)

    def __check_row(self, sec_code, row_code, row):
        '''Итерация по ожидаемым спецификам с их последующей проверкой'''
        specs_map = self.__get_specs(sec_code)
        for col_code, spec_idx in specs_map.items():
            self.__check_format((sec_code, row_code, col_code),
                                SpecInspector, row, spec_idx, specs_map)

    def __check_cells(self, sec_code, row_code, row):
        '''Итерация по значениям строки с их последующей проверкой'''
        for col_code, value in row.items():
            self.__check_format((sec_code, row_code, col_code),
                                ValueInspector, value)

    def __check_format(self, coords, inspector_class, *args):
        '''Инициализация инспектора, проверка'''
        inspector = inspector_class(self.__get_format(*coords),
                                    self._schema.catalogs)
        inspector.check(coords, *args)

    def __get_format(self, sec_code, row_code, col_code):
        '''Возвращает словарь с условиями проверки'''
        try:
            return self._schema.formats[sec_code][row_code][col_code]
        except KeyError:
            raise NoRuleError(sec_code, row_code, col_code)

    def __get_specs(self, sec_code):
        '''Возвращает словарь со спецификами'''
        try:
            return self._schema.formats[sec_code]['specs']
        except KeyError:
            raise NoSectionTemplateError(sec_code)



import re
from ..base import AbstractValidator

year_pattern = re.compile(r'18\d{2}|19\d{2}|20\d{2}')


class AttrValidator(AbstractValidator):
    name = 'Проверка атрибутов'
    code = '1'

    def __init__(self, schema):
        self._schema = schema
        self.errors = []

    def __repr__(self):
        return '<AttrValidator errors={errors}>'.format(**self.__dict__)

    def validate(self, report):
        self._check_year(report)
        self._check_match(report)
        self._check_period(report)

        return not bool(self.errors)

    def _check_year(self, report):
        '''Проверка формата года'''
        if not year_pattern.match(report.year):
            self.error('Указан недопустимый год', '1')

    def _check_match(self, report):
        '''Проверка совпадения типа периода отчёта с периодом схемы'''
        if report.period_type is not None:
            if report.period_type != self._schema.idp:
                self.error('Тип периодичности отчёта не соответствует '
                           'типу периодичности шаблона', '2')

    def _check_period(self, report):
        '''Проверка кода периода'''
        if report.period_code is None:
            if not report.set_periods(self._schema.catalogs, self._schema.idp):
                self.error('Неверное значение периода отчёта', '3')



import traceback
from collections import defaultdict
from .helpers import SchemaFormats, NestedDefaultdict
from .validators import (AttrValidator, TitleValidator,
                         FormatValidator, ControlValidator)


def str_int(v):
    return str(int(v)) if v.isdigit() else v


class Schema:
    def __init__(self, xml_tree, *, skip_warns):
        self.xml = xml_tree
        self.errors = []
        self.required = []
        self.skip_warns = skip_warns
        self.dimension = defaultdict(list)

        self.idp = self._get_idp()
        self.obj = self._get_obj()
        self.title = self._get_title()
        self.formats = self._get_formats()
        self.controls = self._get_controls()
        self.catalogs = self._get_catalogs()

        self.validators = self._init_validators()

    def __repr__(self):
        return '<Schema idp={idp} obj={obj} title={title}'.format(
            **self.__dict__)

    def _get_idp(self):
        '''Получение атрибута idp'''
        return str(int(self.xml.xpath('/metaForm/@idp')[0]))

    def _get_obj(self):
        '''Получение атрибута obj'''
        return self.xml.xpath('/metaForm/@obj')[0]

    def _get_title(self):
        '''Получение ноды с заголовком'''
        return self.xml.xpath('/metaForm/title')[0]

    def _get_formats(self):
        '''Итерация по секциям, строкам и колонокам с получением нод,
           определяющих формат строк и значений в отчёте
        '''
        form = SchemaFormats()
        for section in self.xml.xpath('/metaForm/sections/section'):
            sec_code = str_int(section.attrib['code'])
            defaults, specs = self.__get_default_formats(section, sec_code)
            form[sec_code] = {'specs': specs}

            for row in section.xpath('./rows/row[@type!="C"]'):
                row_code = str_int(row.attrib['code'])
                form[sec_code][row_code] = defaults.copy()

                for cell in row.xpath('./cell'):
                    col_code = str_int(cell.attrib['column'])
                    form[sec_code][row_code][col_code] = cell.attrib

                    if self.__required_cell(row, cell):
                        coords = (sec_code, row_code, col_code)
                        self.required.append(coords)
        return form

    def __required_cell(self, row, cell):
        '''Проверка, является ли ячейка обязательной к заполнению'''
        if cell.attrib['inputType'] == '1' and row.attrib['type'] != 'M':
            return True
        return False

    def __get_default_formats(self, section, sec_code):
        '''Получение нод определяющих формат строк и значений по умолчанию'''
        defaults, specs = {}, {}
        for column in section.xpath('./columns/column'):
            col_code = str_int(column.attrib['code'])

            if column.attrib['type'] == 'S':
                specs[col_code] = column.attrib['fld'][-1]
            elif column.attrib['type'] == 'Z':
                defaults[col_code] = self.__get_default_cell(column)
                self.dimension[sec_code].append(col_code)

        return defaults, specs

    def __get_default_cell(self, column):
        '''Возвращает словарь атрибутов дефолтной ячейки или пустой словарь'''
        try:
            return column.find('default-cell').attrib
        except AttributeError:
            return {}

    def _get_controls(self):
        '''Получение нод с контролями'''
        return self.xml.xpath('/metaForm/controls/control')

    def _get_catalogs(self):
        '''Получение справочников'''
        catalogs = defaultdict(dict)
        for catalog in self.xml.xpath('/metaForm/dics/dic'):
            catalog_id = catalog.attrib['id']
            catalogs[catalog_id]['full'] = NestedDefaultdict(set)
            catalogs[catalog_id]['ids'] = []

            for term_node in catalog.xpath('./term'):
                term_id = term_node.attrib.pop('id')

                catalogs[catalog_id]['ids'].append(term_id)

                for attr, value in term_node.attrib.items():
                    catalogs[catalog_id]['full'][term_id][attr].add(value)
        return catalogs

    def _init_validators(self):
        '''Инициализация валидаторов'''
        return (AttrValidator(self),
                TitleValidator(self),
                FormatValidator(self),
                ControlValidator(self))

    def validate(self, report):
        '''Валидация отчёта'''
        try:
            for validator in self.validators:
                if not validator.validate(report):
                    self._errors_handle(validator)
                    break
        except Exception:
            self.errors.append({'code': '0.0',
                                'name': 'Непредвиденная ошибка',
                                'description': 'Не удалось выполнить проверку',
                                'level': 0})
            print('Unexpected Error', traceback.format_exc())
        finally:
            return self.errors

    def _errors_handle(self, validator):
        '''Форматирование ошибок'''
        for error in validator.errors:
            self.errors.append({
                'code': f'{validator.code}.{error.code}',
                'name': validator.name,
                'description': error.description,
                'level': error.level
            })



import re
from ..exceptions import PeriodExprError

in_pattern = re.compile(r'^\(&npin\(([\d,]+)\)\)$')
cp_pattern = re.compile(r'^&np([=<>]+)(\d+)$')


class PeriodInspector:
    def __init__(self, control):
        self.period_clause = control.attrib.get('periodClause', '').strip()

    def __repr__(self):
        return '<PeriodInspector clause={period_clause}>'.format(
            **self.__dict__)

    def _normolize_period_clause(self):
        '''Нормализация правила'''
        self.period_clause = self.period_clause.replace(' ', '').lower()
        self.period_clause = self.period_clause.replace('and', ' and ')
        self.period_clause = self.period_clause.replace('or', ' or ')

    def check(self, report):
        if not self.period_clause:
            return True

        self._normolize_period_clause()

        if 'in' in self.period_clause:
            return self._check_in(report)
        else:
            return self._check_logic(report)

    def _eval_regex(self, pattern, string):
        '''Разбор формулы проверки периода с помощью регулярки'''
        result = pattern.match(string)
        if result is None:
            raise PeriodExprError()
        return result

    def _to_int(self, clause_parts):
        return ','.join(str(int(n)) for n in clause_parts.split(','))

    def _check_in(self, report):
        '''Проверка на вхождение в список'''
        clause_parts = self._eval_regex(in_pattern, self.period_clause)

        clause = '{0} in ({1},)'.format(report.period_code,
                                        self._to_int(clause_parts.group(1)))
        return eval(clause)

    def _check_logic(self, report):
        '''Проверка комплексного логического условия'''
        clause_parts = []
        for cluase_part in self.period_clause.strip('()').split():
            if cluase_part in ('or', 'and'):
                clause_parts.append(cluase_part)
                continue

            op, num = self._eval_regex(cp_pattern, cluase_part).group(1, 2)
            op = '==' if op == '=' else op
            clause_parts.append(f'{report.period_code} {op} {num}')

        clause = ' '.join(clause_parts)
        return eval(clause)



import ply.yacc as yacc
from .lexer import tokens
from .elements import ElemLogic, ElemSelector, ElemList, Elem

precedence = (
    ('left', 'LOGIC'),
    ('left', 'COMP'),
    ('left', 'PLUS', 'MINUS'),
    ('left', 'MULTIPLY', 'DIVIDE'),
    ('left', 'SUM', 'ABS'),
    ('right', 'UMINUS'),            # Unary minus operator
)

op_name = {
    '+': 'add',
    '-': 'sub',
    '*': 'mul',
    '/': 'truediv',
}


def p_elem_logic(p):
    '''elem : elem COMP elem
            | elem LOGIC elem'''
    p[0] = ElemLogic(p[1], p[2].strip('| '), p[3])


def p_elem_selector(p):
    '''elem : COALESCE elems
            | NULLIF elems'''
    p[0] = ElemSelector(p[1].lower(), p[2])


def p_elem_func(p):
    '''elem : ABS elem
            | SUM elem
            | FLOOR elem'''
    p[0] = p[2]
    p[0].add_func(p[1].lower(), None)


def p_elem_func_args(p):
    '''elem : ISNULL LPAREN elems RPAREN
            | ROUND LPAREN elems RPAREN'''
    p[0], *args = p[3]
    p[0].add_func(p[1].lower(), *args)


def p_elem_func_sum(p):
    '''elem : SUM LPAREN elem RPAREN'''
    p[0] = p[3]
    p[0].add_func(p[1].lower(), None)


def p_elem_math(p):
    '''elem : elem PLUS elem
            | elem MINUS elem
            | elem MULTIPLY elem
            | elem DIVIDE elem'''
    p[1].add_func(op_name[p[2]], p[3])
    p[0] = p[1]


def p_elem(p):
    '''elem : ELEM_START coords ELEM_END'''
    p[0] = ElemList(*p[2])


def p_num_elem(p):
    '''elem : NUM'''
    p[0] = Elem(p[1])


def p_coords(p):
    '''coords : CODE
              | coords CODE'''
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[1].append(p[2])
        p[0] = p[1]


def p_elem_group(p):
    '''elems : elem COMMA elem
             | elems COMMA elem'''
    if isinstance(p[1], list):
        p[1].append(p[3])
        p[0] = p[1]
    else:
        p[0] = [p[1], p[3]]


def p_ne_elem(p):
    '''elem : MINUS elem %prec UMINUS'''
    p[0] = -p[2]


def p_elem_in_parens(p):
    '''elem : LPAREN elem RPAREN
       elems : LPAREN elems RPAREN'''
    p[0] = p[2]


def p_error(p):
    print('Unexpected token:', p)


parser = yacc.yacc()



from setuptools import setup, find_packages

setup(
    name='rosstat-flc',
    version='1.1.1',
    packages=find_packages(),
    description='Tool for format-logistic control of reports sent to RosStat',
    long_description=open('README.md', 'r').read(),
    long_description_content_type="text/markdown",
    author='Nikita Ryabinin',
    author_email='ryabinin.ne@gmail.com',
    license='MIT',
    license_file='LICENSE',
    url='https://github.com/WoolenSweater/rosstat_flc',
    install_requires=['lxml', 'ply'],
    python_requires='>=3.7',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ]
)



from math import gcd
from typing import List, Dict, Tuple, Optional
from collections import defaultdict as defdict
from dataclasses import dataclass, InitVar, field as f
from lxml.etree import _ElementTree
from .schema import str_int


def max_divider(num, terms):
    '''НОД для списка чисел'''
    for term_id in terms:
        num = gcd(num, int(term_id))
    return num


@dataclass
class Row:
    code: str
    s1: str
    s2: str
    s3: str
    cols: Dict[str, str] = f(default_factory=dict)
    _blank: bool = True

    @property
    def blank(self):
        '''Флаг указывающий, что строка пустая'''
        return self._blank

    def items(self, codes=None):
        '''Итерация по элементам строки'''
        codes = codes or self.cols.keys()
        for col_code in codes:
            yield col_code, self.get_col(col_code)

    def get_col(self, code):
        '''Возвращает указанную колонку'''
        return self.cols.get(code)

    def add_col(self, col_code, col_text):
        '''Добавление колонки в строку'''
        self.cols[col_code] = col_text
        self._blank = False

    def get_spec(self, idx):
        '''Возвращает специфику по её "индексу"'''
        return getattr(self, f's{idx}')


@dataclass
class Section:
    code: str
    rows: List[Row] = f(default_factory=list)
    row_codes: List[str] = f(default_factory=list)

    _ignore_specs: Tuple[set] = f(default=({None}, {'*'}, {'0'}), repr=False)
    _ignore_report_specs: Tuple[str] = f(default=('XX',), repr=False)

    def items(self, codes=None, specs=None):
        '''Итерация по элементам раздела'''
        codes = codes or sorted(set(self.row_codes), key=int)
        for row_code in codes:
            yield row_code, list(self.get_rows(row_code, specs=specs))

    def get_rows(self, code, specs=None):
        '''Возвращает строки с указанным кодом и спецификами'''
        for row_code, row in zip(self.row_codes, self.rows):
            if row_code == code and self._check_specs(row, specs):
                yield row

    def _check_specs(self, row, specs):
        '''Проверка, входит ли строка в список переданных специфик'''
        if specs is None:
            return True
        for i in range(1, 4):
            row_spec = getattr(row, f's{i}')
            if row_spec in self._ignore_report_specs:
                return True
            if specs[i] not in self._ignore_specs and row_spec not in specs[i]:
                return False
        return True

    def add_row(self, row_code, row):
        '''Добавление строки в раздел'''
        self.row_codes.append(row_code)
        self.rows.append(row)


@dataclass
class Report:
    xml: InitVar[_ElementTree]
    _blank: bool = True
    _year: str = None
    _title: Dict[str, str] = None
    _data: Dict[str, Section] = None
    _period_raw: str = None
    _period_type: Optional[str] = None
    _period_code: Optional[str] = None
    _row_counters: defdict = f(default_factory=lambda: defdict(int))

    def __repr__(self):
        return '<Report title={_title}\ndata={_data}>'.format(**self.__dict__)

    def __post_init__(self, xml):
        self._title = self._read_title(xml)
        self._data = self._read_data(xml)

        self._get_periods(xml)
        self._get_year(xml)

    @property
    def year(self):
        return self._year

    @property
    def blank(self):
        return self._blank

    @property
    def title(self):
        return self._title

    @property
    def period_type(self):
        return self._period_type

    @property
    def period_code(self):
        return self._period_code

    @property
    def row_counters(self):
        return self._row_counters

    def items(self):
        '''Итерация по разделам отчёта'''
        for sec_code, section in self._data.items():
            yield sec_code, section

    def get_section(self, section_code):
        '''Возвращает указанную секцию'''
        return self._data.get(section_code)

    def _read_title(self, xml):
        '''Чтение заголовка отчёта'''
        title = []
        for node in xml.xpath('/report/title/item'):
            item = (node.attrib['name'], node.attrib.get('value', '').strip())
            title.append(item)
        return title

    def _read_data(self, xml):
        '''Чтение тела отчёта (разделы/строки/колонки)'''
        data = {}
        for section_xml in xml.xpath('/report/sections/section'):
            section_code = str_int(section_xml.attrib['code'])
            section = Section(section_code)

            for row_xml in section_xml.xpath('./row'):
                row_code = str_int(row_xml.attrib['code'])
                row = Row(row_code, *self._read_row_specs(row_xml))

                for col in row_xml.xpath('./col'):
                    col_code = str_int(col.attrib['code'])

                    row.add_col(col_code, col.text)
                    self._blank = False
                section.add_row(row_code, row)
                self._row_counters[(row_code, row.s1, row.s2, row.s3)] += 1
            data[section_code] = section
        return data

    def _read_row_specs(self, row):
        '''Чтение спицифик строки'''
        return (row.attrib.get(f's{i}') for i in range(1, 4))

    def _get_year(self, xml):
        '''Получение года из корня отчёта'''
        self._year = xml.xpath('/report/@year')[0]

    def _get_periods(self, xml):
        '''Получение и разбиение периода из корня отчёта'''
        self._period_raw = xml.xpath('/report/@period')[0]
        if len(self._period_raw) == 4:
            self._period_type = str_int(self._period_raw[:2])
            self._period_code = str_int(self._period_raw[2:])

    def set_periods(self, catalogs, idp):
        '''Попытка привести тип и код периода к формату
           описанному в приказе Росстата
        '''
        try:
            periods_id = self._get_periods_id(catalogs)

            if int(self._period_raw) not in periods_id:
                return False

            max_code = max(periods_id)

            if max_code <= int(idp):
                self._period_type = idp
                self._period_code = self._period_raw
                return True

            max_div = max_divider(max_code, periods_id)

            if max_code <= int(idp) * max_div:
                self._period_type = idp
                self._period_code = str(int(int(self._period_raw) / max_div))
                return True
            return False
        except Exception:
            return False

    def _get_periods_id(self, catalogs):
        '''Получение идентификаторов допустимых периодов из справочника'''
        try:
            return [int(term_id) for term_id in catalogs['s_time']['ids']]
        except KeyError:
            return [int(term_id) for term_id in catalogs['s_mes']['ids']]



from collections import namedtuple

Error = namedtuple('Error', ('description', 'code', 'level'))


class AbstractValidator:
    def __init__(self, schema):
        self._schema = schema

    def error(self, *args, level=1):
        self.errors.append(Error(*args, level))

    def validate(self, report):
        raise NotImplementedError



from ..exceptions import (ValueBaseError, ValueNotNumberError, ValueBadFormat,
                          ValueNotInRangeError, ValueNotInListError,
                          ValueNotInDictError, ValueLengthError)


class ValueInspector:
    def __init__(self, params, catalogs):
        self._catalogs = catalogs

        self.catalog = params.get('dic')
        self.format = params.get('format')
        self.vld_type = params.get('vldType')
        self.vld_param = params.get('vld')

        self.format_funcs_map = {'N': self._is_num, 'C': self._is_chars}

    def __repr__(self):
        return ('<ValueInspector format={format} vld_type={vld_type} '
                'vld_param={vld_param}>').format(**self.__dict__)

    @classmethod
    def _is_num(self, value, limits):
        '''Проверка длины целой и дробной частей числового значения поля'''
        try:
            float(value)
        except ValueError:
            raise ValueNotNumberError()

        value_parts = tuple(len(n) for n in value.split('.'))
        if len(value_parts) == 1:
            i_part_len, f_part_len = value_parts[0], 0
        else:
            i_part_len, f_part_len = value_parts[0], value_parts[1]
        i_part_lim, f_part_lim = (int(n) for n in limits.split(','))

        if not (i_part_len <= i_part_lim and f_part_len <= f_part_lim):
            raise ValueBadFormat()

    @classmethod
    def _is_chars(self, value, limit):
        '''Проверка длины символьного значения поля'''
        if not len(value) <= int(limit):
            raise ValueLengthError()

    def check(self, coords, value):
        try:
            self.__check_format(value)
            self.__check_value(value)
        except ValueBaseError as ex:
            ex.update(coords)
            raise

    def __check_format(self, value):
        '''Разбор "формулы" проверки формата. Вызов метода проверки'''
        alias, args = self.format.strip(' )').split('(')
        format_check_func = self.format_funcs_map[alias]
        format_check_func(value, args)

    def __check_value(self, value):
        if self.vld_type == '1':
            self.__check_value_catalog(value)
        elif self.vld_type == '2':
            self.__check_value_range(value)
        elif self.vld_type == '3':
            self.__check_value_list(value)

    def __check_value_catalog(self, value):
        '''Проверка на вхождение в справочник'''
        if value not in self._catalogs[self.catalog]['ids']:
            raise ValueNotInDictError()

    def __check_value_range(self, value):
        '''Проверка на вхождение в диапазон'''
        value = float(value)
        start, end = (int(n) for n in self.vld_param.split('-'))
        if not (value >= start and value <= end):
            raise ValueNotInRangeError()

    def __check_value_list(self, value):
        '''Проверка на вхождение в список'''
        if value not in self.vld_param.split(','):
            raise ValueNotInListError()
